<?php
//class to define vertex
class vertex{
	private $label;
	private $visited;
	
	function __construct($label){
		$this->label = $label;
		$this->visited = false;
	}
	
	function getLabel(){
		return $this->label;
	}
	
	function visited(){
		$this->visited = true;
	}
	
	function isVisited(){
		return $this->visited;
	}
}	
	
//class to define edge
class edge{
	private $start_vertex;
	private $end_vertex;
	private $cost;
	
	function __construct($vertex1,$vertex2,$cost){
		$this->start_vertex = new vertex($vertex1);
		$this->end_vertex = new vertex($vertex2);
		$this->cost = $cost;
	}
	
	function getStartVertex(){ 
		return $this->start_vertex;
	}
	
	function getEndVertex(){ 
		return $this->end_vertex;
	}
	
	function getCost(){
		return $this->cost;
	}
}

/*
class to define Minimum Spanning tree. constructMST constructs the minimum spanning tree using heap.
formCluster forms clusters by deleting the most expensive edge. BreadthFisrtSearch is used to traverse the MST
*/ 
class tree {
	private $cluster_heap;
	private $vertices;
	private $adjMatrix;
	
	function __construct(){
		$this->cluster_heap = new cluster();
		$this->vertices = array();
	} 
	
	function constructMST($edges){
		foreach($edges as $edge){
			$this->cluster_heap->insert($edge);
			$vertex1 = $edge->getStartVertex();
			$vertex2 = $edge->getEndVertex();
			$this->adjMatrix[$vertex1->getLabel()][$vertex2->getLabel()] = $vertex2;
			$this->adjMatrix[$vertex2->getLabel()][$vertex1->getLabel()] = $vertex1;
			if(empty($this->vertices) || !in_array($vertex1,$this->vertices)) 
			$this->vertices[$vertex1->getLabel()] = $vertex1;
			if(empty($this->vertices) || !in_array($vertex2,$this->vertices)) 
			$this->vertices[$vertex2->getLabel()] = $vertex2;
		}
		
	}

	
	function formCluster($k){
		$this->cluster_heap->top();
		$nodeQueue = new queue();
		for($j = 0; $j<$k-1; $j++){
			$max_edge = $this->cluster_heap->extract();
			$cluster1_start = $max_edge->getStartVertex()->getLabel();
			$cluster2_start = $max_edge->getEndVertex()->getLabel();
			$this->adjMatrix[$cluster1_start][$cluster2_start] = -1;
			$this->adjMatrix[$cluster2_start][$cluster1_start] = -1;
			$nodeQueue->enqueue($max_edge->getStartVertex());
			$nodeQueue->enqueue($max_edge->getEndVertex());
		}
		$queue = new queue();
		$i=0;
		while(!$nodeQueue->isEmpty()){
			$node = $nodeQueue->dequeue();
			if($this->vertices[$node->getLabel()]->isVisited() == false){
				$this->vertices[$node->getLabel()]->visited();
				$cluster[$i][] = $this->vertices[$node->getLabel()]->getLabel();
				$queue->enqueue($this->vertices[$node->getLabel()]);
				while(!$queue->isEmpty()){
					$node = $queue->dequeue();
					while(($nextnode = $this->getNextVertex($node)) != -1){
						$this->vertices[$nextnode->getLabel()]->visited();
						$cluster[$i] []= $this->vertices[$nextnode->getLabel()]->getLabel();
						$queue->enqueue($this->vertices[$nextnode->getLabel()]);
					}
				}
			}
		$i++;
		}
	return $cluster;
	}

	function getNextVertex($vertex){
		foreach($this->adjMatrix[$vertex->getLabel()] as $vert=>$value){
			if($value != -1 && $this->vertices[$value->getLabel()]->isVisited() == false)
				return $this->adjMatrix[$vertex->getLabel()][$vert];
			
		}
		return -1;
	}
	
	function displayClusters($clusters){
		$k =1;
		foreach($clusters as $cluster){
			print("Cluster ".$k."={");
			for($i=0; $i<count($cluster); $i++)
				if($i != count($cluster)-1)
					print($cluster[$i].",");
				else
					print($cluster[$i]);
				print("}<br/>");
				$k++;
		}
	}
}
//heap to maintain the MST
class cluster extends SplHeap{

	public function compare($edge1,$edge2){
		$values1 = $edge1->getCost();
        $values2 = $edge2->getCost();
        if ($values1 == $values2) return 0;
        return $values1 < $values2 ? -1 : 1;
	}
}
//heap to maintain the tree
class tree_cluster extends SplHeap{

	public function compare($edge1,$edge2){
		$values1 = $edge1->getCost();
        $values2 = $edge2->getCost();
        if ($values1 == $values2) return 0;
        return $values1 > $values2 ? -1 : 1;
	}
}

//queue for the BFS traversal
class queue{
	private $size = 20;
	private $queArray;
	private $front;
	private $rear;
	
	function __construct(){
		$this->queArray = array();
		$this->front = 0;
		$this->rear = -1;
	}
	
	function enqueue($i){
		if($this->rear == $this->size-1)
			$this->rear = -1;
		$this->queArray[++$this->rear] = $i;
	}
	
	function dequeue(){
		$temp = $this->queArray[$this->front++];
		if($this->front == $this->size)
			$this->front = 0;
		return $temp;
	}
	function isEmpty(){
		if(($this->rear+1)== $this->front || ($this->front+$this->size-1) == $this->rear)
			return true;
		return false;
	}
	
}
// function to create tree from the input and apply kruskal's algorithm to find MST.
function construct_tree($edges){
	$vertices = array();
	$tree_heap = new tree_cluster();
	$vertice_no = 1;
	for($i=0; $i<count($edges)-1;$i++){
		$edge1 = new edge($edges[$i][0],$edges[$i][1],$edges[$i][2]);
		$tree_heap->insert($edge1);
		$vertex1 = $edge1->getStartVertex();
		$vertex2 = $edge1->getEndVertex();
		if(empty($vertices[$vertex1->getLabel()])){
				$vertices[$vertex1->getLabel()] = $vertice_no;
				$vertice_no++;
		}
		if(empty($vertices[$vertex2->getLabel()])){
				$vertices[$vertex2->getLabel()] = $vertice_no;
				$vertice_no++;
		}
	}
	$k = 0;
	$tree_heap->top();
	while($k < count($vertices)-1){
		$min_edge = $tree_heap->extract();
		$vertex1= $min_edge->getStartVertex()->getLabel();
		$vertex2 = $min_edge->getEndVertex()->getLabel();
		if($vertices[$vertex1]!= $vertices[$vertex2]){
			if($vertices[$vertex1]< $vertices[$vertex2]){
					$m = $vertices[$vertex1];
					$n = $vertices[$vertex2];
			}
			else{
				$m = $vertices[$vertex1];
				$n = $vertices[$vertex2];
			}
			foreach($vertices as $vertex => $no){
				if($no == $m){
					$vertices[$vertex] = $n;
				}
			}
			$min_edges[] = $min_edge;
			$k++;
		}
	}
	return $min_edges;
}
//input tree. # of clusters is 3 which is the input to the function call formCluster.
$edges = array(array("A","B","14"),array("B","C","12"),array("B","D","28"),array("C","D","19"),
			   array("D","E","17"),array("A","F","11"),array("F","C","19"),array("A","E","10"), 
			   array("A","D","30"),array("B","E","30"));
$mst_edges = construct_tree($edges);
$mst = new tree();
$mst->constructMST($mst_edges);
$clusters = $mst->formCluster(3);
print("Data:<br/>");
foreach($edges as $edge)
	print($edge[0].$edge[1]."=>".$edge[2]."<br/>");
print("<br/>Minimum Edges:<br/>");
for($i=0;$i<=count($mst_edges)-1; $i++)
	print($mst_edges[$i]->getStartVertex()->getLabel().$mst_edges[$i]->getEndVertex()->getLabel()."=>".$mst_edges[$i]->getCost()."<br/>");

print("<br/>Clusters:<br/>");
$mst->displayClusters($clusters);
?>